from pathlib import Path
from typing import List
import re, sys
import pandas as pd
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.patches import Patch
import matplotlib.gridspec as gridspec
import seaborn as sns

class EddyFigure(object):
    """Eddy Figure class"""
    def save_all_outlier_slices(self, fig_outdir:str):
        # load mask, pre and post-replace data
        self.load_data()

        self.fig_outdir = Path(fig_outdir)
        self.fig_outdir.mkdir(exist_ok=True)

        # plot them
        for v, s, std, sqr_std, r in zip(self.outlier_vol,
                                         self.outlier_slice,
                                         self.stds,
                                         self.sqr_stds,
                                         self.rank):
            bvalue = self.bvalue_arr[v]
            pre_data_tmp = self.pre_data[:, :, s, v]
            pre_data_tmp = np.where(self.mask_data[:, :, s] == 1,
                                    pre_data_tmp, 0)
            post_data_tmp = self.post_data[:, :, s, v]
            post_data_tmp = np.where(self.mask_data[:, :, s] == 1,
                                     post_data_tmp, 0)

            sagittal_data = self.pre_data[self.mid_point, :, :, v].T
            sagittal_data_fixed = self.post_data[self.mid_point, :, :, v].T

            outfile = self.fig_outdir / f'{r:03}_vol_{v}_slice_{s}.png'

            plot_pre_post_correction_slice(
                self.eddy_dir,
                pre_data_tmp, post_data_tmp,
                sagittal_data, sagittal_data_fixed,
                outfile,
                s, v, bvalue, r,
                std, sqr_std, self.outlier_std_array,
                self.restricted_movement_array)


def plot_pre_post_correction_slice(
        subject: Path,
        pre_data: np.array, post_data: np.array,
        sagittal_data: np.array, sagittal_data_fixed: np.array,
        outfile: Path,
        slice_number: float, volume_number: int, bvalue: float, rank: int,
        outlier_std: float, outlier_sqr_std: float,
        std_array: np.array, motion_array: np.array):
    '''Plot pre vsnd post correction slices

    Key Arguments:
        subject : eddy output directory
        pre_data : numpy array of data for the slice, pre-repol
        post_data : numpy array of data for the slice, post-repol
        sagittal_data : numpy array of data for the slice, pre-repol
        sagittal_data_fixed : numpy array of data for the slice, post-repol
        outfile : output png file location, Path
        slice_number : outlier slice number, int
        volume_number: outlier volume number, int
        bvalue: bvalue, float
        rank: rank of the outlier volume, int
        std_array: array of the deviation from gaussian model, np.array
        motion_array: array of the motion (restricted), np.array

    To do:
        set the vmax and vmin equal for both axes
    '''
    fig = plt.figure(constrained_layout=True,
                     figsize=(15, 10))
    gs0 = gridspec.GridSpec(5, 6, figure=fig)

    # three graphs at the top
    pre_ax = fig.add_subplot(gs0[0:3, 0:2])
    post_ax = fig.add_subplot(gs0[0:3, 2:4])
    diff_ax = fig.add_subplot(gs0[0:3, 4:6])

    # std and motion graphs
    etc_ax = fig.add_subplot(gs0[3, :5])
    motion_ax = fig.add_subplot(gs0[4, :5])

    # two sagittal slices - to show signal drops
    sagittal_ax = fig.add_subplot(gs0[3, 5])
    sagittal_fixed_ax = fig.add_subplot(gs0[4, 5])
    sagittal_ax.set_axis_off()

    # std graph
    add_eddy_std_array_to_ax(etc_ax, std_array, slice_number, volume_number)

    # motion
    add_motion_graph_to_ax(motion_ax, motion_array, volume_number,
                           etc_ax.get_xlim())

    # two sagital slices
    add_sagital_slices_to_ax(sagittal_ax, sagittal_fixed_ax,
                             sagittal_data,
                             sagittal_data_fixed,
                             slice_number)

    diff_ax_img = add_pre_vs_post_eddy_slices(pre_ax, post_ax, diff_ax,
                                              pre_data, post_data)

    # colorbar to the diff map
    # [left, bottom, width, height]
    cbar_ax = fig.add_axes([0.95, 0.3746, 0.02, 0.797-0.3746])
    fig.colorbar(diff_ax_img, cax=cbar_ax)
    # diff_ax.set_title('sqrt(diff_map^2)')

    fig.subplots_adjust(left=0.05,
                        right=0.95,
                        bottom=0.07,
                        top=0.80)

    fig.suptitle(
        f'{subject}\n'
        f'Rank by sqr_stds: {rank} Bvalue: {bvalue:.0f} ' \
        f'Volume: {volume_number} Slice: {slice_number}\n'
        f'std: {outlier_std:.2f}, sqr_std: {outlier_sqr_std:.2f}',
        y=0.97, fontsize=15)

    #plt.tight_layout()
    fig.savefig(outfile, dpi=fig.dpi)
    plt.close()


def add_eddy_std_array_to_ax(etc_ax:plt.Axes, std_array:np.array,
                             slice_number:int, volume_number:int) -> None:
    # str matrix
    # graph at the bottom
    all_std_img = etc_ax.imshow(std_array.T,
                                aspect='auto',
                                origin='lower')
    etc_ax.axhline(slice_number, linestyle='--', c='r')
    etc_ax.axvline(volume_number, linestyle='--', c='r')

    etc_ax.set_title(
        'Eddy std array (std of mean difference '
        'between observation and prediction)')
    etc_ax.set_ylabel('Slices')
    etc_ax.set_xticks([])


def add_motion_graph_to_ax(motion_ax:plt.Axes, motion_array:np.array,
                           volume_number:int, xlim:np.array) -> None:
    motion_ax.plot(motion_array[:, 0], label='Absolute')
    motion_ax.plot(motion_array[:, 1], label='Relative')
    motion_ax.axvline(volume_number, linestyle='--', c='r')
    motion_ax.set_title('Restricted motion')
    motion_ax.set_xlabel('Volumes')
    motion_ax.set_ylabel('Motion')
    motion_ax.set_xlim(xlim)
    legend = motion_ax.legend(loc='upper left')


def add_sagital_slices_to_ax(sagittal_ax:plt.Axes,
                             sagittal_fixed_ax:plt.Axes,
                             sagittal_data:np.array,
                             sagittal_data_fixed:np.array,
                             slice_number:int) -> None:

    # slice pointer
    sagittal_ax.set_title('Sagittal slice')
    sagittal_ax.imshow(sagittal_data, cmap='gray', origin='lower',
                       aspect='auto')
    val = 3
    sagittal_ax.annotate('',
                         xy=(val, slice_number),
                         xytext=(0, slice_number),
                         arrowprops=dict(facecolor='white',
                                         shrink=0.05),)

    sagittal_ax.annotate('',
                         xy=(sagittal_data.shape[1]-val, slice_number),
                         xytext=(sagittal_data.shape[1], slice_number),
                         arrowprops=dict(facecolor='white',
                                         shrink=0.05),)

    sagittal_fixed_ax.set_title('Sagittal slice (post-eddy)')
    sagittal_fixed_ax.imshow(sagittal_data_fixed,
                             cmap='gray', origin='lower',
                             aspect='auto')
    val = 3
    sagittal_fixed_ax.annotate('',
                               xy=(val, slice_number),
                               xytext=(0, slice_number),
                               arrowprops=dict(facecolor='white',
                                               shrink=0.05),)

    sagittal_fixed_ax.annotate('',
                               xy=(sagittal_data_fixed.shape[1]-val,
                                   slice_number),
                               xytext=(sagittal_data_fixed.shape[1],
                                       slice_number),
                               arrowprops=dict(facecolor='white',
                                               shrink=0.05),)


def add_pre_vs_post_eddy_slices(pre_ax:plt.Axes,
                                post_ax:plt.Axes,
                                diff_ax:plt.Axes,
                                pre_data:np.array,
                                post_data:np.array):
    post_ax_img = post_ax.imshow(post_data, cmap='gray')
    post_ax.set_title('After Eddy outlier replacement\n'
                      '(before motion correction)')
    post_ax_clim = post_ax_img.get_clim()

    pre_ax_img = pre_ax.imshow(pre_data,
                               cmap='gray',
                               vmin=post_ax_clim[0],
                               vmax=post_ax_clim[1])
    pre_ax.set_title('Before Eddy outlier replacement')

    diff_map = np.sqrt((post_data - pre_data)**2)
    diff_ax_img = diff_ax.imshow(diff_map)

    diff_ax.set_title('√(post_data-pre_data)²')

    for ax in post_ax, pre_ax, diff_ax:
        ax.set_axis_off()

    return diff_ax_img


def outlier_df_plot(outlier_df, std_threshold=3):
    '''
    Plot outlier_df

    Args:
        outlier_df: pandas dataframe, containing outlier information.
        std_threshold: integer, to be multiplied to the
                       standard deviation to select outliers.
                       Default:3.

    Returns:
        fig, ax : matplotlib figure and axes.

    To do later:
        - control width according to the size of the input df.
    '''
    map_dict = {'outlier': 1,
                'almost outlier': 0.5,
                '-': 0}
    arr = np.array(outlier_df.applymap(lambda x: map_dict[x]))
    arr = np.where(arr == 0, np.nan, arr)

    fig, ax = plt.subplots(ncols=1, figsize=(20, 10))
    ax.imshow(arr, cmap='Reds', vmax=1, vmin=0)

    cmap = matplotlib.cm.get_cmap('Reds')
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1)

    legend_elements = [
        Patch(facecolor=cmap(norm(1)), label='Outlier'),
        Patch(facecolor=cmap(norm(0.5)), label='Almost outlier')
    ]
    ax.legend(handles=legend_elements,
              loc='center left',
              bbox_to_anchor=(1, 0.5))

    ax.set_yticks(np.arange(len(outlier_df)))
    ax.set_yticklabels(outlier_df.index)

    ax.set_xticks(np.arange(len(outlier_df.columns)))
    ax.set_xticklabels(outlier_df.columns, rotation=90)

    fig.suptitle(
        'Excluded (Threshold : mean + {} std)\n{} subjects'.format(
            std_threshold,
            len(outlier_df)),
        y=.94,
        fontsize=15
    )

    return fig, ax


def motion_summary_figure(study_eddy_runs, std_threshold=3):
    '''
    Draw line graphs of various motion measures from FSL eddy.

    Args:
        study_eddy_runs: list of EddyRun object.
        std_threshold: number to multiply by standard deviation to mark
                       outlier, int.

    Return:
        fig, axes, df
    '''

    movement_df = get_movement_df(study_eddy_runs)
    fig, axes = plt.subplots(ncols=3, nrows=2, figsize=(20, 20))

    # for each data column in movement_df
    for ax, data, title in zip(
        np.ravel(axes),
        [movement_df[x] for x in movement_df.columns if x != 'eddy_path'],
        [x for x in movement_df.columns if x != 'eddy_path']):

        # line graph for data
        ax.plot(data)

        # Data into array
        arr = np.array(data)

        # Get average motion for all subject
        avg = np.mean(data)

        # Draw an average line
        # and legend to include the number of total subjects
        ax.axhline(y=avg, color='g', alpha=0.8, linestyle='--',
                   label='mean :{}'.format(len(np.array(data))))

        # Get standard deviation for all subject motion
        std = np.std(data)

        # Draw 2*std line
        ax.axhline(y=avg+2*std, color='r', alpha=0.3, linestyle='--',
                   label='mean + 2*std :{}'.format(
                       len(arr[arr>(avg+2*std)])
                   ))

        # Draw 3*std line
        ax.axhline(y=avg+3*std, color='r', alpha=0.9, linestyle='--',
                   label='mean + 3*std :{}'.format(
                       len(arr[arr>(avg+3*std)])
                   ))

        ax.set_title(title)
        ax.legend()

        if title.split(' ')[-1] in ['method', 'relative']:
            ax.set_ylim(0,2)
        ax.set_xlabel('Subject')
        ax.set_ylabel(title)

        if title.startswith('Number'):
            ax.set_ylabel('Number of outlier slice')

    return fig, axes


def shell_alignment_summary(list_of_eddy_runs, std_threshold=3):
    #     list_of_eddy_runs = study_eddy_runs
    df = pd.DataFrame()
    for x in list_of_eddy_runs:
        df_tmp = x.post_eddy_shell_alignment_df
        df_tmp['subject'] = x.ep
        df = pd.concat([df, df_tmp])
        for col in df.columns[:6]:
            df[col] = df[col].astype(float)

    df_long_all =  pd.melt(
            df,
            id_vars=['subtitle', 'title', 'subject', 'shell_info'],
            var_name='motion_parameter',
            value_name='motion')
    df_long_all['axis'] = df_long_all['motion_parameter'].str[0]
    df_long_all['motion_type'] = \
        df_long_all['motion_parameter'].str.extract('(tr|rot)')


    for subtitle_unique, df_for_subtitle in df_long_all.groupby('subtitle'):
        for title_unique, df_for_subtitle_title in df_for_subtitle.groupby('title'):
            # axes
            gb = df_for_subtitle_title.groupby(['shell_info', 'motion_type'])
            ncols = len(df_for_subtitle_title['shell_info'].unique())
            nrows = len(df_for_subtitle_title['motion_type'].unique())
            fig, axes = plt.subplots(
                    ncols=ncols,
                    nrows=nrows,
                    figsize=(ncols*10, nrows*3))

            for ax, (gb_id, df_tmp) in zip(np.ravel(axes, order='F'), gb):
                for axis, table in df_tmp.groupby('axis'):
                    table = table.set_index('subject')
                    table = table.reindex(list(df_long_all.subject.unique()))
                    ax.plot(table['motion'].to_list(), label=axis)

                ax.set_title('\n'.join(gb_id))
                ax.set_xlabel('subject number')
                ax.legend()
            fig.suptitle(title_unique+'\n'+subtitle_unique, y=1.1, fontsize=20)
            fig.tight_layout()

    return fig, axes

def shell_PE_translation_summary(list_of_eddy_runs, std_threshold=3):
    df = pd.DataFrame()
    for x in list_of_eddy_runs:
        df_tmp = x.post_eddy_shell_PE_translation_parameters_df
        df_tmp['subject'] = x.ep
        df = pd.concat([df, df_tmp])

    df[0] = df[0].str.split(' ').str[0].astype(float)
    df.columns = ['motion', 'shell_info', 'title', 'subtitle', 'subject']

    for subtitle_unique, df_for_subtitle in df.groupby('subtitle'):
        for title_unique, df_for_subtitle_title in df_for_subtitle.groupby('title'):
            # axes
            gb = df_for_subtitle_title.groupby('shell_info')
            nrows = len(df_for_subtitle_title['shell_info'].unique())
            fig, axes = plt.subplots(nrows=nrows, figsize=(5*nrows, 10))

            for ax, (gb_id, df_tmp) in zip(np.ravel(axes, order='F'), gb):
                df_tmp = df_tmp.set_index('subject')
                df_tmp = df_tmp.reindex(list(df.subject.unique()))
                ax.plot(df_tmp['motion'].to_list(), label=gb_id)

                ax.set_title(gb_id)
                ax.set_xlabel('subject number')
            fig.suptitle(title_unique+'\n'+subtitle_unique, y=1.1, fontsize=20)
            fig.tight_layout()
            fig.show()


def motion_summary_dist(study_eddy_runs, std_threshold=3):

    fig, axes = plt.subplots(ncols=3, nrows=2, figsize=(20, 20))

    outlier_df = pd.DataFrame(columns=['subjects'])

    for ax, data, title in zip(np.ravel(axes), 
                              [[x.ofer_movement_avg for x in study_eddy_runs],
                               [x.movement_avg[1] for x in study_eddy_runs],
                               [x.restricted_movement_avg[1] for x in study_eddy_runs],
                               [x.movement_avg[0] for x in study_eddy_runs],
                               [x.restricted_movement_avg[0] for x in study_eddy_runs],
                               [x.number_of_outlier_slices for x in study_eddy_runs]],
                              ['Ofer motion method', 'Movement : relative', 'Restricted Movement : relative',
                              'Movement : absolute', 'Restricted Movement : absolute', 'Number of outlier slice']):
        ax.hist(data, bins=100)
        avg = np.mean(data)
        std = np.std(data)

        ax.axvline(x=avg, color='g', alpha=0.8, linestyle='--', label='mean')
        ax.axvline(x=avg+2*std, color='r', alpha=0.3, linestyle='--', label='mean + 2*std')
        ax.axvline(x=avg+3*std, color='r', alpha=0.9, linestyle='--', label='mean + 3*std')


        std_threshold = 3
        outlier_data_map = np.where(np.array(data) >  avg+std_threshold*std)
        outlier_eddy_run = list(study_eddy_runs[i] for i in outlier_data_map[0])
        outlier_df_tmp = pd.DataFrame([x.ep for x in outlier_eddy_run], columns=['subjects'])
        outlier_df_tmp[title] = 1

        outlier_df = pd.merge(outlier_df, 
                              outlier_df_tmp,
                              on='subjects', how='outer')

        ax.set_title(title)
        ax.legend()


        ax.set_xlabel(title)
        ax.set_ylabel('Subject count')

        if title.startswith('Number'):
            ax.set_ylabel('Number of outlier slice')

    fig.show()
    outlier_df = outlier_df.fillna(0)


    fig, ax = plt.subplots(ncols=1, figsize=(10, 10))
    ax.imshow(outlier_df.fillna(0).set_index('subjects'))
    ax.set_yticks(np.arange(len(outlier_df)))
    ax.set_yticklabels(outlier_df.subjects)#.str.extract('eddy\/(\d{2}_S\S+)_eddy_corrected')[0])
    ax.set_xticks(np.arange(len(outlier_df.columns[1:])))
    ax.set_xticklabels(outlier_df.columns[1:], rotation=90)
    fig.suptitle('Excluded (Threshold : mean + {} std)\n{} subjects'.format(std_threshold,
                                                                           len(outlier_df)), y=.94, fontsize=15)

    outlier_df['subjects'] = outlier_df.subjects.str.extract('eddy\/(\d{2}_S\S+)_eddy_corrected')
    fig.show()



class EddyStudyFigures:
    '''
    Args:
        study_dir: str, glob input like patterns for eddy directories
        eg) /data/pnl/kcho/*eddy
    '''
    def set_group_figure_settings(self):
        self.dpi = 100

    def plot_subjects(self, var, std_outlier=2):
        '''Create graphs for motion, outlier and etc for all eddy outputs'''
        # width for one number of subjects
        width_per_subject = 0.5

        g = sns.catplot(x='subject', y=var, data=self.df)

        if len(self.df) < 10:
            g.fig.set_size_inches(5, 5)
        else:
            g.fig.set_size_inches(width_per_subject * len(self.df), 5)
            g.ax.set_xticklabels(g.ax.get_xticklabels(), rotation=90)

        g.fig.set_dpi(self.dpi)
        g.ax.set_xlabel('Subjects')

        g.fig.tight_layout()
        g.fig.suptitle(f'{var[0].upper()}{var[1:]}')

        threshold = self.df[var].mean() + (self.df[var].std() * std_outlier)
        g.ax.axhline(y=threshold, color='red', alpha=0.4)
        g.ax.text(x=len(self.df) - 0.5,
                  y=threshold+0.1,
                  s=f'mean + std * {std_outlier}',
                  ha='right', color='red', alpha=0.9)

        setattr(self, f'plot_{var}', g)

        # plot outliers only
        df_tmp = self.df[self.df[var] > threshold]
        x_size = len(df_tmp) * 1.2

        if len(df_tmp) > 0:
            g = sns.catplot(x='subject', y=var, data=df_tmp)
            g.fig.set_size_inches(x_size, 5)
            g.fig.set_dpi(self.dpi)
            g.ax.set_xticklabels(g.ax.get_xticklabels())
            g.ax.set_xlabel('Subjects')

            g.fig.suptitle(f'Subjects with greater {var[0].lower()}{var[1:]} '
                           'than (mean + 2*std)', y=1.02)
            setattr(self, f'plot_outlier_only_{var}', g)


    def create_group_figures(self, out_dir):
        '''Create group figures'''
        self.set_group_figure_settings()

        out_dir = Path(out_dir)
        # save figures
        vars = ['absolute restricted movement', 'relative restricted movement',
                'number of outlier slices',
                'Sum of standard deviations in outlier slices',
                'Mean of standard deviations in outlier slices',
                'Standard deviation of standard deviations in outlier slices']
        for var in vars:
            self.plot_subjects(var)
            g = getattr(self, f'plot_{var}')
            g.fig.savefig(out_dir / f'plot_{var}.png')

            # if there is outlier
            try:
                g = getattr(self, f'plot_outlier_only_{var}')
                g.fig.savefig(out_dir / f'plot_outlier_only_{var}.png')
            except:
                pass

        # Eddy shell alignment
        self.figure_post_eddy_shell()
        post_eddy_shell_graph_list = [x for x in dir(self)
                                      if 'plot_post_eddy_shell' in x]
        for post_eddy_shell_graph in post_eddy_shell_graph_list:
            fig = getattr(self, post_eddy_shell_graph)
            fig.savefig(out_dir / f'{post_eddy_shell_graph}.png')


        self.figure_post_eddy_shell_PE()
        post_eddy_shell_PE_graph_list = [x for x in dir(self)
                                         if 'plot_post_eddy_shell_PE' in x]
        for post_eddy_shell_PE_graph in post_eddy_shell_PE_graph_list:
            fig = getattr(self, post_eddy_shell_PE_graph)
            fig.savefig(out_dir / f'{post_eddy_shell_PE_graph}.png')

        # dataframe clean up
        self.df = self.df.sort_values(
            ['number of outlier slices',
             'Sum of standard deviations in outlier slices',
             'absolute restricted movement',
             'relative restricted movement'], ascending=False).drop(
                 ['eddy_prefix', 'eddy_dir', 'eddy_input'],
                 axis=1).reset_index().drop('index', axis=1)

        self.df.to_csv(out_dir / 'eddy_study_summary.csv')


    def figure_post_eddy_shell_PE(self):
        for (title, subtitle), table in \
            self.post_eddy_shell_PE_translation_parameters_df.groupby(
                ['title', 'subtitle']):
            self.figure_post_eddy_shell_PE_suptitle(title, subtitle, table)

    def figure_post_eddy_shell(self):
        for (title, subtitle), table in \
            self.post_eddy_shell_alignment_df.groupby(
                ['title', 'subtitle']):
            self.figure_post_eddy_shell_suptitle(title, subtitle, table)

    def figure_post_eddy_shell_suptitle(self, title, subtitle, t):
        shell_infos = t['shell_info'].unique()
        if len(t) > 30:
            fig, axes = plt.subplots(
                ncols=len(shell_infos),
                figsize=(3.2*len(shell_infos), 2*len(shell_infos)),
                dpi=self.dpi)
        else:
            fig, axes = plt.subplots(
                ncols=len(shell_infos),
                figsize=(10, 10),
                dpi=self.dpi)

        for ax, shell_info in zip(np.ravel(axes), shell_infos):
            t_tmp = t.groupby('shell_info').get_group(shell_info)
            t_tmp = t_tmp.reset_index()

            std = t_tmp['sum'].std()
            mean = t_tmp['sum'].mean()

            ax.plot(np.arange(len(t_tmp)), t_tmp['sum'], 'ro',
                    alpha=0.3, label='z-rot (deg)')
            ax.axhline(y=mean+2*std, color='r', alpha=0.4)
            ax.text(x=len(t_tmp), y=mean+2*std, s='mean + 2 * std',
                    ha='right', va='top', color='r', alpha=0.4)

            outlier_df = t_tmp[t_tmp['sum'] > (mean + 2*std)]
            for num, row in outlier_df.iterrows():
                ax.text(num, row['sum'],
                        row['subject'],
                        ha='center', fontsize=7)
            print(' '.join(outlier_df.subject.tolist()))
            ax.set_title(shell_info)

        try:
            axes[0].set_ylabel('Sum of xyz translations and rotations')
        except:
            axes.set_ylabel('Sum of xyz translations and rotations')
        fig.suptitle(f'{title}\n{subtitle}', y=1)

        setattr(self, f'plot_post_eddy_shell_{subtitle}', fig)
        # fig.show()

    def figure_post_eddy_shell_PE_suptitle(self, title, subtitle, t):
        shell_infos = t['shell_info'].unique()
        if len(t) > 30:
            fig, axes = plt.subplots(
                ncols=len(shell_infos),
                figsize=(3.2*len(shell_infos), 2*len(shell_infos)),
                dpi=self.dpi)
        else:
            fig, axes = plt.subplots(
                ncols=len(shell_infos),
                figsize=(10, 10),
                dpi=self.dpi)

        for ax, shell_info in zip(np.ravel(axes), shell_infos):
            t_tmp = t.groupby('shell_info').get_group(shell_info)
            t_tmp = t_tmp.reset_index()

            std = t_tmp[0].std()
            mean = t_tmp[0].mean()

            ax.plot(np.arange(len(t_tmp)), t_tmp[0], 'ro',
                    alpha=0.3, label='z-rot (deg)')
            ax.axhline(y=mean+2*std, color='r', alpha=0.4)
            ax.text(x=len(t_tmp), y=mean+2*std, s='mean + 2 * std',
                    ha='right', va='top', color='r', alpha=0.4)

            outlier_df = t_tmp[t_tmp[0] > (mean + 2*std)]
            for num, row in outlier_df.iterrows():
                ax.text(num, row[0],
                        row['subject'],
                        ha='center', fontsize=7)
            print(' '.join(outlier_df.subject.tolist()))
            ax.set_title(shell_info)

        try:
            axes[0].set_ylabel('PE translation in mm')
        except:
            axes.set_ylabel('PE translation in mm')
        fig.suptitle(f'{title}\n{subtitle}', y=1)
        setattr(self, f'plot_post_eddy_shell_PE_{subtitle}', fig)
        # fig.show()



from pathlib import Path
import os
import numpy as np
import sys
import pytest

script_loc = Path(os.path.realpath(__file__))
script_dir = script_loc.parent
test_root_dir = script_dir.parent
# root_dir = test_root_dir.parent
print(test_root_dir)
print(test_root_dir)
sys.path.append(str(test_root_dir))

from eddy_squeeze.eddy_squeeze import eddy_squeeze_study
from eddy_squeeze.eddy_squeeze_lib.eddy_files import EddyRun
from eddy_squeeze.eddy_squeeze_lib.eddy_web import create_study_html
from eddy_squeeze.eddy_squeeze_lib.eddy_files import EddyDirectories


class FakeArgs():
    def __init__(self):
        self.out_dir = None
        self.eddy_dir = None
        self.print_only = None
        self.save_csv = None
        self.save_html = None


@pytest.fixture
def eddyRun():
    eddy_prefix = script_dir / 'prac_out' / 'subject01-eddy_out'
    eddyRun = EddyRun(eddy_prefix)
    # eddyRun.read_file_locations_from_command()
    # eddyRun.load_eddy_information()

    return eddyRun


def test_study_create_html(eddyRun):
    fakeArgs = FakeArgs()
    fakeArgs.eddy_prefix = str(eddyRun.eddy_prefix)
    fakeArgs.out_dir = eddyRun.eddy_prefix.parent / 'prac'
    fakeArgs.save_html = True

    eddy_dirs = [
            (script_dir / 'prac_study_dir' / 'subject01'),
            (script_dir / 'prac_study_dir' / 'subject02'),
            (script_dir / 'prac_study_dir' / 'subject03'),
            ]
    # eddyDirectories = EddyDirectories([str(x) for x in eddy_dirs])
    eddyDirectories = EddyDirectories(eddy_dirs)
    create_study_html(eddyDirectories, fakeArgs.out_dir)


def test_eddy_squeeze_study():
    eddy_dirs = [
            (script_dir / 'prac_study_dir' / 'subject01'),
            (script_dir / 'prac_study_dir' / 'subject02'),
            (script_dir / 'prac_study_dir' / 'subject03'),
            ]
    # eddyDirectories = EddyDirectories([str(x) for x in eddy_dirs])
    eddyDirectories = EddyDirectories(eddy_dirs)
    eddyDirectories.create_group_figures(out_dir='eddy_summary')

    # print('Saving the summary outputs to {args.out}')
    # eddyDirectories.df.to_csv('prac_eddy_study_summary.csv')
    create_study_html(eddyDirectories, out_dir='eddy_summary')


def test_eddy_squeeze_study_print_table():
    eddy_dirs = [
            (script_dir / 'prac_study_dir' / 'subject01'),
            (script_dir / 'prac_study_dir' / 'subject02'),
            (script_dir / 'prac_study_dir' / 'subject03'),
            ]
    fakeArgs = FakeArgs()

    fakeArgs.eddy_prefix_pattern = False
    fakeArgs.print_table = True
    fakeArgs.save_html = False
    fakeArgs.eddy_directories = eddy_dirs
    fakeArgs.figures = None

    eddy_squeeze_study(fakeArgs)


def test_eddy_squeeze_study_html():
    eddy_dirs = [
            (script_dir / 'prac_study_dir' / 'subject01'),
            (script_dir / 'prac_study_dir' / 'subject02'),
            (script_dir / 'prac_study_dir' / 'subject03'),
            ]
    fakeArgs = FakeArgs()

    fakeArgs.eddy_prefix_pattern = False
    fakeArgs.print_table = False
    fakeArgs.save_html = True
    fakeArgs.eddy_directories = eddy_dirs
    fakeArgs.figures = None

    eddy_squeeze_study(fakeArgs)

def test_eddy_squeeze_study_html_with_figure():
    eddy_dirs = [
            (script_dir / 'prac_study_dir' / 'subject01'),
            (script_dir / 'prac_study_dir' / 'subject02'),
            (script_dir / 'prac_study_dir' / 'subject03'),
            ]
    fakeArgs = FakeArgs()

    fakeArgs.eddy_prefix_pattern = False
    fakeArgs.print_table = False
    fakeArgs.save_html = True
    fakeArgs.eddy_directories = eddy_dirs
    fakeArgs.figures = True

    eddy_squeeze_study(fakeArgs)


def test_eddy_squeeze_study_html_with_prefix():
    eddy_prefix = str(script_dir / 'prac_study_dir' / 'subj*')
    fakeArgs = FakeArgs()

    fakeArgs.eddy_prefix_pattern = eddy_prefix
    fakeArgs.print_table = False
    fakeArgs.save_html = True
    fakeArgs.eddy_directories = None
    fakeArgs.figures = None

    eddy_squeeze_study(fakeArgs)




from pathlib import Path
from typing import List
import re, sys
import pandas as pd
import numpy as np

from eddy_squeeze.eddy_squeeze_lib.eddy_collect import EddyOut, EddyCollect
from eddy_squeeze.eddy_squeeze_lib.eddy_present import EddyFigure, EddyStudyFigures
from eddy_squeeze.eddy_squeeze_lib.eddy_utils import get_paths_with_suffixes
from eddy_squeeze.eddy_squeeze_lib.eddy_utils import get_absolute_when_there_are_dots_in_the_paths
from eddy_squeeze.eddy_squeeze_lib.eddy_web import create_html



class EddyRun(EddyOut, EddyCollect, EddyFigure):
    '''EddyRun to catch user given eddy prefix 
       to return object with eddy info'''
    def __init__(self, eddy_prefix:str, **kwargs):
        '''EddyRun initialization

        Key Arguments:
            - eddy_prefix: Eddy output prefix, str
                  eg. '/eddy/out/dir/*eddy_out'

        kwargs arguments:
            - name: name of the data or subject
            - dwi, bval, mask: paths of raw dwi, bval and mask.
                               Giving these paths will overwrite the locatios
                               defined in the eddy command.txt. And will be
                               used to create eddy repol difference figures
        '''

        self.eddy_prefix = Path(eddy_prefix)
        self.eddy_dir = self.eddy_prefix.absolute().parent
        self.eddy_exist = check_if_eddy_ran(self.eddy_dir)

        if 'name' in kwargs:
            self.subject_name = kwargs.get('name')
        else:
            self.subject_name = \
                get_subject_name_from_eddy_prefix(self.eddy_prefix)

        register_eddy_paths(self)

    def read_file_locations_from_command(self, **kwargs):
        '''Register raw inputs to the eddy'''

        # register nifti_input, bval and mask to self
        # based on the file paths in command_txt
        paths = return_paths_from_eddy_command_txt(self.command_txt)
        self.nifti_input, self.bvalue_txt, self.mask = paths


        # this is required for loading pre-repol and post-repol
        # overwrite paths if it's given through kwargs
        if 'dwi' in kwargs:
            self.nifti_input = Path(kwargs.get('dwi'))
        if 'bval' in kwargs:
            self.bvalue_txt = Path(kwargs.get('dwi'))
        if 'mask' in kwargs:
            self.mask = Path(kwargs.get('dwi'))

        self.bvalue_arr = np.loadtxt(str(self.bvalue_txt))

    # def read_and_register_raw_files(self):
        # '''Read eddy command.txt file to load raw data information'''

        # # If there is command_txt, get raw dwi input 
        # # the the eddy from command_txt
        # paths = return_paths_from_eddy_command_txt(self.command_txt)
        # self.nifti_input, self.bvalue_txt, self.mask = paths

        # self.bvalue_arr = np.loadtxt(self.bvalue_txt)


class EddyDirectory(EddyRun):
    def __init__(self, eddy_dir, **kwargs):
        dot_eddy_files = get_dot_eddy_files(eddy_dir)
        self.ep = get_unique_eddy_prefixes(eddy_dir)
        EddyRun.__init__(self, self.ep, **kwargs)


def get_dot_eddy_files(eddy_dir: str) -> List[Path]:
    '''Returns list of paths, that includes .eddy'''
    eddy_files_and_dirs = list(Path(eddy_dir).glob('*.eddy*'))
    dot_eddy_files = [str(x).split('.eddy')[0] for x in eddy_files_and_dirs if Path(x).is_file()]

    if len(dot_eddy_files) == 0:
        print(f'There is no eddy related files in {eddy_dir}')

    return dot_eddy_files


def get_unique_eddy_prefixes(dot_eddy_files: List[Path]) -> str:
    '''
    Returns list of unique eddy prefixes

    Args:
        eddy_dir: string or Path object of eddy out directory
    '''
    dot_eddy_prefix_unique = set(dot_eddy_files)

    if len(dot_eddy_prefix_unique) == 1:
        return dot_eddy_prefix_unique.pop()
    elif len(dot_eddy_prefix_unique) > 1:
        print('There are more than two unique eddy prefix')
        print(f'\t{dot_eddy_prefix_unique}')
        return dot_eddy_prefix_unique.pop()
    else:
        sys.exit(f'There is no eddy related files')


def is_there_any_eddy_output(paths: List[Path]):
    '''Return True if there is any path that matches eddy output'''
    paths_including_eddy_in_fname = [x for x in paths if '.eddy' in x.name]

    if len(paths_including_eddy_in_fname) > 1:
        return True
    else:
        return False


def check_if_eddy_ran(user_given_dir:str):
    '''Return True if there are eddy outputs'''
    user_given_dir = Path(user_given_dir)

    paths_with_eddy_in_name = user_given_dir.glob('*.eddy*')
    is_eddy_ran = is_there_any_eddy_output(paths_with_eddy_in_name)

    return is_eddy_ran


def get_subject_name_from_eddy_prefix(eddy_prefix:Path, pattern=None):
    '''Return subject name from eddy prefix'''

    # to absolute path
    eddy_prefix = eddy_prefix.absolute()

    if pattern is not None:
        subject_name = re.search(pattern, str(eddy_prefix.parent)).group(1)
    else:
        subject_name = eddy_prefix.parent.name

    return subject_name


def get_eddy_files(eddy_prefix) -> dict:
    '''
    Returns dictionary of eddy output file locations in string

    Args:
        EddyRun
        eddy_prefix: string of eddy prefix

    '''
    eddy_postfixes = ['post_eddy_shell_alignment_parameters',
                      'outlier_map',
                      'rotated_bvecs',
                      'outlier_n_stdev_map',
                      'outlier_n_sqr_stdev_map',
                      'outlier_report',
                      'parameters',
                      'restricted_movement_rms',
                      'post_eddy_shell_PE_translation_parameters',
                      'values_of_all_input_parameters',
                      'command_txt',
                      'movement_rms']

    eddy_files_dict = {}
    for eddy_postfix in eddy_postfixes:
        eddy_files_dict[eddy_postfix] = eddy_prefix.with_suffix(
            f'.eddy_{eddy_postfix}')

    eddy_files_dict['eddy_out_data'] = eddy_prefix.with_suffix('.nii.gz')
    eddy_files_dict['outlier_free_data'] = eddy_prefix.with_suffix(
        '.eddy_outlier_free_data.nii.gz')

    return eddy_files_dict


def register_eddy_paths(eddyRun:EddyRun) -> None:
    '''Register eddy out files to attributes

    Key arguments:
        - eddyRun
    '''
    eddy_files_dict = get_eddy_files(eddyRun.eddy_prefix)
    for name, file_loc in eddy_files_dict.items():
        setattr(eddyRun, name, file_loc)


def return_paths_from_eddy_command_txt(command_txt_loc:Path) -> tuple:
    """Read eddy command.txt file to load raw data information"""

    with open(command_txt_loc, 'r') as f:
        command = f.read()

    nifti_input = Path(re.search(r'--imain=(\S+)', command).group(1))
    bvalue_txt = Path(re.search(r'--bvals=(\S+)',command).group(1))
    mask = Path(re.search(r'--mask=(\S+)', command).group(1))

    nifti_input, mask = get_paths_with_suffixes([nifti_input, mask])

    nifti_input, bvalue_txt, mask = \
        get_absolute_when_there_are_dots_in_the_paths(
            [nifti_input, bvalue_txt, mask], command_txt_loc.parent)

    return (nifti_input, bvalue_txt, mask)



class EddyDirectories(EddyStudyFigures):
    def __init__(self, eddy_dirs:List[Path], **kwargs):
        self.eddy_dirs = eddy_dirs
        print(f'Summarizing {len(self.eddy_dirs)} subjects')

        self.study_eddy_runs = []
        self.eddy_dir_error = []
        self.eddy_prefix_list = []
        self.eddyRuns = []
        for eddy_dir in self.eddy_dirs:
            try:
                dot_eddy_files = get_dot_eddy_files(eddy_dir)
                eddy_dir_ep = get_unique_eddy_prefixes(dot_eddy_files)
                self.eddy_prefix_list.append(eddy_dir_ep)

                if 'name' in kwargs:
                    eddyRun = EddyRun(eddy_dir_ep, name=eddy_dir)
                else:
                    eddyRun = EddyRun(eddy_dir_ep)

                eddyRun.read_file_locations_from_command()
                eddyRun.load_eddy_information()
                eddyRun.get_outlier_info()
                eddyRun.estimate_eddy_information()
                eddyRun.outlier_summary_df()

                self.eddyRuns.append(eddyRun)

            except SystemExit:
                pass
                # pass
        # self.df = pd.concat([x.df.to_frame() for x in self.eddyRuns], axis=1).T
        self.df = pd.concat([x.df for x in self.eddyRuns], axis=1).T
        self.create_sub_df_for_each_information()

        self.df_motion = pd.concat([x.df_motion for x in self.eddyRuns])

        self.post_eddy_shell_alignment_df = pd.concat(
            [x.post_eddy_shell_alignment_df for x in self.eddyRuns],
            axis=0)
        self.post_eddy_shell_PE_translation_parameters_df = pd.concat(
            [x.post_eddy_shell_PE_translation_parameters_df for x
                in self.eddyRuns],
            axis=0)

    def save_all_outlier_slices(self):
        '''Run all_outlier_slices for all eddyRun objects'''
        for eddyRun in self.eddyRuns:
            fig_outdir = eddyRun.eddy_dir / 'eddy_squeeze_qc'
            eddyRun.save_all_outlier_slices(fig_outdir)

    def save_all_html(self):
        '''Run all_outlier_slices for all eddyRun objects'''
        for eddyRun in self.eddyRuns:
            create_html(eddyRun, out_dir=eddyRun.eddy_dir)

    def get_unique_bvalues(self):
        unique_b_values = np.stack(self.df['unique b values'])
        return np.unique(unique_b_values, axis=0)

    def create_sub_df_for_each_information(self):
        '''Divide self.df into different dfs for better visualization'''

        self.subdf_basics = self.df[
            ['subject', 'eddy_dir', 'number of volumes', 'max b value',
             'min b value', 'unique b values', 'number of b0s']]

        self.subdf_outliers = self.df[
            ['subject', 'number of outlier slices',
             'Sum of standard deviations in outlier slices',
             'Mean of standard deviations in outlier slices',
             'Standard deviation of standard deviations in outlier slices']]

        self.subdf_motions = self.df[
            ['subject', 'absolute restricted movement',
             'relative restricted movement']]

    def clean_up_data_frame(self):
        pass
        # if 'name_set' in self.df.columns:
            # self.df.index = self.df.name_set
        # else:
            # self.df.index = self.df.ep.apply(
                # lambda x: Path(x).name.split('-eddy_out')[0]).to_list()

        # self.df.index.name = 'subject'
        # self.df = self.df.reset_index()

        # self.post_eddy_shell_alignment_df.index = \
            # self.post_eddy_shell_alignment_df.subject

        # self.post_eddy_shell_alignment_df.index.name = 'subject'
        # self.post_eddy_shell_alignment_df = \
            # self.post_eddy_shell_alignment_df.reset_index()

        # self.post_eddy_shell_PE_translation_parameters_df.index = \
            # self.post_eddy_shell_PE_translation_parameters_df.ep.apply(
                # lambda x: Path(x).name.split('-eddy_out')[0]).to_list()
        # self.post_eddy_shell_PE_translation_parameters_df.index.name = \
                # 'subject'
        # self.post_eddy_shell_PE_translation_parameters_df = \
            # self.post_eddy_shell_PE_translation_parameters_df.reset_index()




    def get_basic_diff_info(self):
        self.df.groupby(
            ['number of volumes',
             'max b value',
             'min b value',
             'number of b0s']).count()['subject'].to_frame()



import numpy as np
import pandas as pd
import nibabel as nb
import re, warnings


class EddyCollect():
    def __init__(self):
        pass

    def get_outlier_info(self):
        #TODO add bvalue information to the outlier information
        # find a list of outlier slices
        self.outlier_vol, self.outlier_slice = \
            np.where(self.outlier_array == 1)

        # find a list of std in the outlier slices
        self.stds = self.outlier_std_array[
            self.outlier_vol, self.outlier_slice]

        # find a list of sqr of std in the outlier slices
        self.sqr_stds = self.outlier_sqr_std_array[
            self.outlier_vol,
            self.outlier_slice]

        # get bvalues
        self.outlier_bvalues = self.bvalue_arr[self.outlier_vol]

        # get a order of stds
        # get a order of sqr_stds
        self.rank = (-np.absolute(self.sqr_stds)).argsort().argsort()

    def load_data(self):
        # load data
        self.mask_data = nb.load(self.mask).get_fdata()
        self.pre_data = nb.load(self.nifti_input).get_fdata()
        self.post_data = nb.load(self.outlier_free_data).get_fdata()

        # get mid point of data x-axis
        self.mid_point = int(self.post_data.shape[0] / 2)

    def outlier_summary_df(self):
        """Create summary dataframes
        self.df: details of outlier slices
        self.df_motion: average motion parameter
        """
        df = pd.DataFrame()

        df['Volume'] = self.outlier_vol
        df['Slice'] = self.outlier_slice
        df['B value'] = self.outlier_bvalues
        df['Stds'] = self.stds
        df['Sqr_stds'] = self.sqr_stds
        df['rank'] = self.rank
        df['subject'] = self.subject_name

        self.eddy_outlier_df = df.sort_values(by='rank').reset_index().drop(
            'index',
            axis=1)
        self.eddy_outlier_df.index.name = 'Outlier slices'

        df_motion = pd.DataFrame()
        df_motion['restricted_absolute_motion'] = \
            [self.restricted_movement_avg[0]]
        df_motion['restricted_relative_motion'] = \
            [self.restricted_movement_avg[1]]
        for var in [
                'number_of_outlier_slices',
                'outlier_std_total', 'outlier_std_mean', 'outlier_std_std']:
            df_motion[var] = getattr(self, var)
        self.df_motion = df_motion



def estimate_average_motions(movement_rms):
    '''
    Return a tuple of averages (absolute, relative)

    Args:
        movement_rms: string of ep.eddy_movement_rms file
    '''
    mean_motion = np.loadtxt(movement_rms)
    motion_avg = mean_motion.mean(axis=0)
    return motion_avg


def make_df_from_lines_post_eddy_shell_alignment(lines):

    df = pd.DataFrame()

    # Get title and subtitle
    for line_number, line in enumerate(lines):
        if re.search('are|were', line):
            title = line
            if lines[line_number+1].startswith('Shell'):
                subtitle = line
        elif len(re.search('[A-Za-z ]*', line).group(0)) > 10:
            subtitle = line
        elif line.startswith('Shell'):
            shell_info = line
            array_info = lines[line_number+2]
            df_tmp = pd.DataFrame(array_info.split()).T
            df_tmp['shell_info'] = shell_info
            df_tmp['title'] = title
            df_tmp['subtitle'] = subtitle

            df = pd.concat([df, df_tmp])

    df.columns = ['x-tr (mm)', 'y-tr (mm)', 'z-tr (mm)',
                  'x-rot (deg)', 'y-rot (deg)', 'z-rot (deg)',
                  'shell_info',
                  'subtitle',
                  'title']
    return df


def get_post_eddy_shell_alignment_in_df(
        post_eddy_shell_alignment_parameters):
    '''
    Return pandas data frame of post eddy shell alignment
    '''

    with open(post_eddy_shell_alignment_parameters, 'r') as f:
        text = f.read()

    lines = text.split('\n')

    df = make_df_from_lines_post_eddy_shell_alignment(lines)

    for axis in ['x', 'y', 'z']:
        for var in ['-tr (mm)', '-rot (deg)']:
            df[axis+var] = df[axis+var].astype(float).apply(np.absolute)

    df['sum'] = df['x-tr (mm)'] + df['y-tr (mm)'] + df['z-tr (mm)'] + \
        df['x-rot (deg)'] + df['y-rot (deg)'] + df['z-rot (deg)']

    return df


def make_df_from_lines_post_eddy_shell_PE_translation_parameters(lines):
    df = pd.DataFrame()

    # Get title and subtitle
    for line_number, line in enumerate(lines):
        if re.search('are|were', line):
            title = line
            if lines[line_number+1].startswith('Shell'):
                subtitle = line
        elif re.search(r'PE-translations \(mm\)', line):
            subtitle = line
        elif line.startswith('Shell'):
            shell_info = line.split(':')[0]
            try:
                pe_translation = re.search(r'(\S+\.|\S+)\ mm', line).group(1)
                pe_translation = np.absolute(float(pe_translation))
            except:
                pe_translation = ''
            df_tmp = pd.DataFrame([pe_translation])
            df_tmp['shell_info'] = shell_info
            df_tmp['title'] = title
            df_tmp['subtitle'] = subtitle

            df = pd.concat([df, df_tmp])

    return df


def get_post_eddy_shell_PE_translation_parameters(
        post_eddy_shell_PE_translation_parameters):
    '''
    Return pandas data frame of post eddy shell alignment
    '''
    with open(post_eddy_shell_PE_translation_parameters, 'r') as f:
        text = f.read()

    lines = text.split('\n')
    df = make_df_from_lines_post_eddy_shell_PE_translation_parameters(lines)

    return df


def get_movement_df(study_eddy_runs):
    '''
    Reads values from list of study_eddy_run classes.

    Args:
        study_eddy_runs: list of Eddy_run classes. list.

    Returns:
        movement_df: table of information for eacch Eddy_run classes.
                     Pandas dataframe.
    '''

    # Ofer measure is removed - Wednesday, July 31, 2019
    #'Ofer measure': [x.ofer_movement_avg for x in study_eddy_runs],

    movement_df = pd.DataFrame()

    for er in study_eddy_runs:
        tmp_dict = {
            'eddy_path': [er.eddy_prefix],
            'Absolute Restricted Movement': [er.restricted_movement_avg[0]],
            'Relative Restricted Movement': [er.restricted_movement_avg[1]],
            'Absolute Movement': [er.movement_avg[0]],
            'Relative Movement': [er.movement_avg[1]],
            'Number of outlier slice': [er.number_of_outlier_slices]
        }
        movement_df_tmp = pd.DataFrame(tmp_dict)
        movement_df = pd.concat([movement_df, movement_df_tmp])

    return movement_df


def get_outlier_df(study_eddy_runs, std_threshold=3):
    '''
    Define which eddy runs are outliers

    Args:
        study_eddy_runs: list of Eddy_run classes. list.
        std_threshold: integer, to be multiplied to the
                       standard deviation to select outliers.
                       Default:3

    Returns:
        pandas dataframe of outlier information
    '''
    # Load data from each eddy run into a dataframe
    movement_df = get_movement_df(study_eddy_runs)

    # Sort by
    movement_df = movement_df.sort_values([
        'Relative Restricted Movement',
        'Number of outlier slice'],
        ascending=False)

    outlier_df = movement_df.copy()

    # For each measure (columns in the movement_df)
    for col in [x for x in outlier_df.columns if x != 'eddy_path']:
        avg = outlier_df[col].mean()
        std = outlier_df[col].std()

        # Create an array that marks outliers
        # True marks outlier
        outlier_df.loc[
                outlier_df[col] > avg + std * std_threshold,
                col] = 'outlier'

        # False marks "almost outlier"
        # Select top 3 nearest subjects to the threshold
        outlier_df.loc[
            (outlier_df.loc[outlier_df[col] != 'outlier', col].sort_values(
                ascending=False)[:3].index),
            col] = 'almost outlier'

    # Drop lines with no True or False
    outlier_df = outlier_df.set_index('eddy_path')
    outlier_df = outlier_df[
        outlier_df.isin(['outlier', 'almost outlier']).any(axis=1)]

    outlier_df[~outlier_df.isin(['outlier', 'almost outlier'])] = '-'

    return outlier_df


class EddyOut:

    def load_movement_arrays(self):
        '''Load information from movement file path attributes'''
        self.restricted_movement_array = np.loadtxt(
            self.restricted_movement_rms)

        self.movement_array = np.loadtxt(self.movement_rms)

    def load_outlier_arrays(self):
        '''Load outlier information'''
        # Outlier volume, slice, and number
        self.outlier_report_df = get_outlier_report(self.outlier_report)

        # outlier map, stdev, and n_sar_stdev
        self.outlier_array = np.loadtxt(self.outlier_map, skiprows=1)

        # The numbers denote how many standard deviations off the mean
        # difference between observation and prediction is.
        self.outlier_std_array = np.loadtxt(self.outlier_n_stdev_map,
                                            skiprows=1)

        # The numbers denote how many standard deviations off the square root
        # of the mean squared difference between observation and prediction is.
        self.outlier_sqr_std_array = np.loadtxt(self.outlier_n_sqr_stdev_map,
                                                skiprows=1)

    def get_info_movement_arrays(self):
        '''Extract motion summary from the loaded motion arrays'''

        # restricted movement
        self.restricted_movement_avg = estimate_average_motions(
            self.restricted_movement_rms)

        # general movement
        self.movement_avg = estimate_average_motions(self.movement_rms)
        self.absolute_movement_array = self.movement_array[0, :]
        self.relative_movement_array = self.movement_array[1, :]

    def get_info_post_eddy(self):
        '''Post-eddy information into Pandas DataFrame'''
        self.post_eddy_shell_alignment_df = \
            get_post_eddy_shell_alignment_in_df(
                self.post_eddy_shell_alignment_parameters)

        self.post_eddy_shell_PE_translation_parameters_df = \
            get_post_eddy_shell_PE_translation_parameters(
                self.post_eddy_shell_PE_translation_parameters)

    def get_info_outlier_arrays(self):
        '''Extract information from outlier arrays'''

        om_array = np.loadtxt(self.outlier_map, skiprows=1)
        outlier_slice = np.where(om_array == 1)[1]
        outlier_slice_number = len(outlier_slice)
        self.number_of_outlier_slices = outlier_slice_number

        self.outlier_std_total = self.outlier_std_array[
            np.where(self.outlier_array == 1)].sum()
        self.outlier_std_total = np.absolute(self.outlier_std_total)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            self.outlier_std_mean = self.outlier_std_array[
                np.where(self.outlier_array == 1)].mean()

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            self.outlier_std_std = self.outlier_std_array[
                np.where(self.outlier_array == 1)].std()

        self.outlier_std_mean = np.absolute(self.outlier_std_mean)
        self.outlier_std_std = np.absolute(self.outlier_std_std)

    def load_eddy_information(self):
        '''Load eddy information'''
        
        self.volume_in_each_bshell = \
            get_number_of_directions_in_each_shell_dict(self.bvalue_arr)

        self.load_movement_arrays()
        self.load_outlier_arrays()

    def estimate_eddy_information(self):
        self.get_info_movement_arrays()
        self.get_info_outlier_arrays()
        self.get_info_post_eddy()
        # print(tabulate(self.post_eddy_shell_alignment_df, headers='keys', tablefmt='psql'))
        # print(tabulate(self.post_eddy_shell_PE_translation_parameters_df, headers='keys', tablefmt='psql'))
        self.collect_all_info()


    def collect_all_info(self):
        df = pd.Series()

        df['subject'] = self.subject_name

        df['eddy_prefix'] = self.eddy_prefix
        df['eddy_dir'] = self.eddy_dir
        df['eddy_input'] = self.nifti_input

        # bvalue
        df['number of volumes'] = len(self.bvalue_arr)
        df['max b value'] = self.bvalue_arr.max()
        df['min b value'] = self.bvalue_arr.min()
        df['unique b values'] = np.unique(self.bvalue_arr)
        df['number of b0s'] = len(self.bvalue_arr[self.bvalue_arr == 0])

        # outlier information
        df['number of outlier slices'] = self.number_of_outlier_slices
        df['Sum of standard deviations in outlier slices'] = \
            self.outlier_std_total
        df['Mean of standard deviations in outlier slices'] = \
            self.outlier_std_mean
        df['Standard deviation of standard deviations in outlier slices'] = \
            self.outlier_std_std

        # movement information
        df['absolute restricted movement'] = self.restricted_movement_avg[0]
        df['relative restricted movement'] = self.restricted_movement_avg[1]
        df['absolute movement'] = self.movement_avg[0]
        df['relative movement'] = self.movement_avg[1]

        self.df = df
        self.post_eddy_shell_alignment_df['eddy_prefix'] = self.eddy_prefix
        self.post_eddy_shell_alignment_df['subject'] = self.subject_name
        self.post_eddy_shell_PE_translation_parameters_df['eddy_prefix'] = self.eddy_prefix
        self.post_eddy_shell_PE_translation_parameters_df['subject'] = \
                self.subject_name


def get_outlier_report(outlier_report):
    '''
    Return pandas data frame of eddy outlier report
    '''
    with open(outlier_report, 'r') as f:
        text = f.read()

    df = pd.DataFrame()
    for i in text.split('\n'):
        try:
            slice_number = re.search(r'Slice (\d+)', i).group(1)
            volume_number = re.search(r'scan (\d+)', i).group(1)
            mean_std = re.search(r'mean (\S*\d+.\d+)', i).group(1)
            mean_squared_std = re.search(
                r'mean squared (\S*\d+.\d+)', i).group(1)

            df_tmp = pd.DataFrame(
                    [slice_number,
                     volume_number,
                     mean_std,
                     mean_squared_std]).T
            df = pd.concat([df, df_tmp])
        except:
            #last line has no info
            pass

    try:
        df.columns = ['slice_num', 'vol_num', 'mean_std', 'mean_squared_std']
    except:
        pass

    return df


def get_number_of_directions_in_each_shell_dict(bvalue_arr):
    '''Return number of directions in each shell'''
    volume_in_each_bshell = {}
    shell, count = np.unique(bvalue_arr, return_counts=True)
    for shell, count in zip(shell, count):
        volume_in_each_bshell[shell] = count
    return volume_in_each_bshell





from pathlib import Path
from typing import List
from tabulate import tabulate
import pandas as pd

def print_table(df:pd.DataFrame) -> None:
    '''Prints df using tabulate'''
    # print(tabulate(df, headers='keys', tablefmt='github'))
    print(tabulate(df, headers='keys', tablefmt='psql'))

def get_paths_with_suffixes(paths:list) -> List[Path]:
    '''Return paths with nifti suffix if missing suffix'''

    new_paths = []

    for path in paths:
        suffixes = path.suffixes
        if len(suffixes) == 0:
            new_paths.append(path.with_suffix('.nii.gz'))
        else:
            new_paths.append(path)

    return new_paths


def get_absolute_paths(paths:list, root:Path) -> List[Path]:
    '''Return paths with nifti suffix if missing suffix'''

    new_paths = []

    for path in paths:
        if path.is_absolute():
            new_paths.append(path)
        else:
            if '..' in str(path):
                pass
            else:
                path_absolute = path.absolute()
                new_paths.append(path_absolute)

    return new_paths


def get_absolute_when_there_are_dots_in_the_path(path:Path, root:Path):
    '''Return absolute path when there is dots in the paths'''
    if str(path).startswith('../'):
        new_root = root.parent
        new_path = Path(str(path)[3:])
        absolute_path = get_absolute_when_there_are_dots_in_the_path(
            new_path, new_root)
    else:
        absolute_path = root / path

    return absolute_path


def get_absolute_when_there_are_dots_in_the_paths(paths:Path, root:Path):
    '''Return absolute paths when there is dots in the paths'''
    new_paths = []

    for path in paths:
        new_paths.append(get_absolute_when_there_are_dots_in_the_path(
            path, root))

    return new_paths


def print_header(header:str) -> None:
    '''Print Header'''
    print()
    print(header)
    print('-'*50)




from pathlib import Path
import os
import argparse
from tabulate import tabulate
# development
import sys
import argparse


from eddy_squeeze.eddy_squeeze_lib.eddy_files import EddyDirectories
from eddy_squeeze.eddy_squeeze_lib.eddy_utils import print_header, print_table
from eddy_squeeze.eddy_squeeze_lib.eddy_web import create_study_html, create_html

from typing import List



def get_all_eddy_prefix_paths(eddy_prefix_pattern:str) -> List[Path]:
    '''Get eddy prefix paths from eddy prefix pattern'''
    if Path(eddy_prefix_pattern).is_absolute(): # absolute path
        # [1:] removes '/'
        eddy_prefix_list = list(Path('/').glob(eddy_prefix_pattern[1:]))
    else: # relative path
        eddy_prefix_list = list(Path('.').glob(eddy_prefix_pattern))

    return eddy_prefix_list


def get_eddy_prefix_list(args: 'argparse') -> List[Path]:
    '''Return eddy_prefix_list based on either a pattern or paths'''
    if args.eddy_prefix_pattern:
        print_header(f'Finding matching eddy prefixes')
        eddy_prefix_list = get_all_eddy_prefix_paths(args.eddy_prefix_pattern)
    elif args.eddy_directories:
        print_header(f'Setting up eddy directories')
        eddy_prefix_list = args.eddy_directories
    else:
        sys.exit('Please provide either '
                 '"--eddy_prefix" or "--eddy_directories"')

    return eddy_prefix_list


def eddy_squeeze_study(args: 'argparse') -> None:
    '''Run eddy_squeeze on a group of Eddy outputs'''

    # set eddy summary output directory
    if not args.out_dir:
        out_dir = Path(os.getcwd()) / 'eddy_summary'
    else:
        out_dir = Path(args.out_dir)
    
    print_header(f'Output directory : {out_dir}')

    # get eddy directory prefixes
    eddy_prefix_list = get_eddy_prefix_list(args)

    # get eddyDirectories
    # 1. read files from the command_txt
    # 2. load movement and outlier information
    # 3. clean up outlier slice information
    # 4. clean up movement information
    # 5. collects all information into eddyDirectories.df
    print_header('Extracting information from all eddy outputs')
    eddyDirectories = EddyDirectories(eddy_prefix_list)
    print_header(f'n={len(eddyDirectories.eddyRuns)} eddy outputs detected')

    # print output
    if args.print_table:
        print_table_from_eddyDirectories(eddyDirectories)

    # create figure
    if args.figures:
        print_header(f'Creating summary figures for all eddy outputs')
        eddyDirectories.create_group_figures(out_dir)

    if args.save_html:
        print_header(f'Creating html summary')
        eddyDirectories.save_all_html()
        create_study_html(eddyDirectories, out_dir=out_dir)


def print_table_from_eddyDirectories(eddyDirectories:EddyDirectories):
    '''Print three separate tables to the shell'''
    print_header('Basic information')
    print_table(eddyDirectories.subdf_basics)

    print_header('Outlier information')
    print_table(eddyDirectories.subdf_outliers)

    print_header('Motion information')
    print_table(eddyDirectories.subdf_motions)



class EddyRunCommand():
    def __init__(self):
        pass


def eddy(echo_spacing, img_in, bvec, bval, mask,
         eddy_out_prefix, repol_on=True):
    '''
    Run FSL eddy

    Parameters:
    echo_spacing (float) : echo spacing of the diffusion data
    other inputs (string)

    https://www.jiscmail.ac.uk/cgi-bin/webadmin?A2=fsl;92bd6f89.1403
        The echo spacing in the Siemens PDF does need to be divided
        by the ipat/GRAPPA factor.  Multi-band has no effect on
        echo spacing.

    https://www.jiscmail.ac.uk/cgi-bin/webadmin?A2=fsl;92136ade.1506
        What I can say is that in _most_ cases it doesn’t really matter
        what you put in the last column. The value is essentially used
        internally to calculate "observed_distortions->estimated_field
        ->estimated_distortions” where the value is used at both -> .
        That means that if you get it wrong, the two errors will cancel.

        It is only really important if
        1. You want your estimated fields to be correctly scaled in Hz
        2. You have acquisitions with different readout times in your
           data set

        Otherwise you can typically just go with 0.05.
        And also make sure that you use the same values for topup and
        eddy if you are using them together.
        Jesper

    https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/eddy/Faq
        There are some special cases where it matters to get the --acqp
        file right, but unless you know exactly what you are doing it
        is generally best to avoid those cases.
        They would be
        - If you acquire data with PE direction along two different
          axes (i.e. the x- and y-axis). In that case you need to get
          the signs right for the columns indicating the PE. But you can
          always use trial and error to find the correct combination.
        - If you acquire data with different readout times. In that case
          you need at the very least to get the ratio between the times
          right.
        - If you use a non-topup derived fieldmap, such as for example
          a dual echo-time gradient echo fieldmap, that you feed into
          eddy as the --field parameter. In this case you need to get
          all signs and times right, both when creating the field (for
          example using prelude) and when specifying its use in eddy
          through the --acqp.

    '''
    eddy_out_dir = dirname(eddy_out_prefix)

    # index
    data_img = nb.load(img_in)
    index_array = np.tile(1, data_img.shape[-1])
    index_loc = join(eddy_out_dir, 'index.txt')
    np.savetxt(index_loc, index_array, fmt='%d', newline=' ')

    # acqp
    acqp_num = (128-1) * echo_spacing * 0.001
    acqp_line = '0 -1 0 {}'.format(acqp_num)
    acqp_loc = join(eddy_out_dir, 'acqp.txt')
    with open(acqp_loc, 'w') as f:
        f.write(acqp_line)

    if repol_on:
        # eddy_command
        eddy_command = '/data/pnl/soft/pnlpipe3/fsl/bin/eddy_openmp \
            --imain={data} \
            --mask={mask} \
            --index={index} \
            --acqp={acqp} \
            --bvecs={bvecs} \
            --bvals={bvals} \
            --repol \
            --out={out}'.format(data=img_in,
                                mask=mask,
                                index=index_loc,
                                acqp=acqp_loc,
                                bvecs=bvec,
                                bvals=bval,
                                out=eddy_out_prefix)
    else:
        # eddy_command
        eddy_command = '/data/pnl/soft/pnlpipe3/fsl/bin/eddy_openmp \
            --imain={data} \
            --mask={mask} \
            --index={index} \
            --acqp={acqp} \
            --bvecs={bvecs} \
            --bvals={bvals} \
            --out={out}'.format(data=img_in,
                                mask=mask,
                                index=index_loc,
                                acqp=acqp_loc,
                                bvecs=bvec,
                                bvals=bval,
                                out=eddy_out_prefix)

    print(re.sub(r'\s+', ' ', eddy_command))
    run(eddy_command)


def eddy_qc(echo_spacing, bvec, bval, mask, eddy_out_prefix):
    '''
    Run FSL eddy qc

    Parameters:
        echo_spacing (float) : echo spacing of the diffusion data
        other inputs (string)

        bvec is not currently used
    '''
    eddy_out_dir = dirname(eddy_out_prefix)
    quad_outdir = '{}.qc'.format(eddy_out_prefix)
    index_loc = join(eddy_out_dir, 'index.txt')
    acqp_loc = join(eddy_out_dir, 'acqp.txt')

    command = f'{config["eddy_quad_location"] {eddy_out_prefix} \
            -idx {index_loc} \
            -par {acqp_loc} \
            -m {mask} \
            -b {bval} \
            -o {quad_outdir}'

    run(command)

def eddy_squad(qc_dirs, outdir):
    try:
        os.mkdir(outdir)
    except:
        pass
    out_text_file = join(outdir, 'eddy_quad_dirs.txt')
    with open(out_text_file, 'w') as f:
        for qc_dir in qc_dirs:
            f.write('{}\n'.format(qc_dir))

    command = f'{config["eddy_squad_location"] {out_text_file} -o {outdir}'
    run(command)



import setuptools
from os.path import join

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements= open('requirements.txt').read().split()

setuptools.setup(
    name="eddy-squeeze", # Replace with your own username
    version="1.0.0",
    author="Kevin Cho",
    author_email="kevincho@bwh.harvard.edu",
    description="Visualize extra information from FSL 6.0.1 eddy outputs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pnlbwh/eddy-squeeze",
    packages=setuptools.find_packages(),
    package_data={'eddy_squeeze':[
        join('html_templates','*'),
        ]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.7',
    install_requires=requirements,
    scripts=['bin/eddy_squeeze']
)



from jinja2 import Environment, FileSystemLoader
from pathlib import Path
import os
import re


root = Path(os.path.abspath(__file__)).parent.parent
static_dir = root.parent / 'docs'
templates_dir = root / 'html_templates'
bwh_fig_loc = templates_dir / 'pnl-bwh-hms.png'

# jinja2 environment settings
env = Environment(loader=FileSystemLoader(str(templates_dir)))


# type
from typing import NewType
EddyStudy = NewType('EddyStudy', object)

def basename(path):
    '''functions used in the jinja2 template'''
    return Path(path).name


def sorter(file_path):
    '''functions used in the jinja2 template'''
    return int(file_path.name[:3])


def create_study_html(eddyStudy:EddyStudy, out_dir:str, **kwargs):
    '''Create html that summarizes eddy directorries'''

    # summary out directory settings
    out_dir = Path(out_dir)
    out_dir.mkdir(exist_ok=True)

    study_out_html = out_dir / 'eddy_study_summary.html'
    print(study_out_html)


    # eddyRun
    env.filters['basename'] = basename
    template = env.get_template('base.html')
    
    html_addresses = []
    for eddyRun in eddyStudy.eddyRuns:
        eddyRun_out_dir = eddyRun.eddy_dir / 'eddy_squeeze_qc'
        image_list = list(sorted(eddyRun_out_dir.glob('*png'), key=sorter))

        out_html = out_dir / \
                f'{eddyRun.subject_name}_eddy_summary.html'
        print(out_html)

        with open(out_html, 'w') as fh:
            fh.write(template.render(image_list=image_list,
                                     eddyOut=eddyRun,
                                     subject=eddyRun.eddy_prefix,
                                     bwh_fig_loc=bwh_fig_loc,
                                     study_out_html=study_out_html
                                     ))

        replace_image_locations_to_relative_in_html(out_html, out_dir)
        html_addresses.append(out_html)

    template = env.get_template('base_study.html')
    # list of images in the output dir created by another function
    image_list = list(out_dir.glob('*png'))
    with open(study_out_html, 'w') as fh:
        print('hahah')
        print(study_out_html)
        fh.write(template.render(out_dir=out_dir,
                                 image_list=image_list,
                                 eddyStudy=eddyStudy,
                                 bwh_fig_loc=bwh_fig_loc,
                                 html_addresses=html_addresses
                                 ))

    replace_image_locations_to_relative_in_html(study_out_html, out_dir)

    # from weasyprint import HTML
    # HTML(study_out_html).write_pdf(out_dir.absolute() / 'Test.pdf')


def create_html(eddyOut, out_dir:str, **kwargs):
    '''Create html that summarizes individual eddy outputs'''

    if 'out_dir' in kwargs:
        out_dir = Path(kwargs.get('out_dir'))
    else:
        out_dir = eddyOut.eddy_dir / 'outlier_figures'

    out_dir.mkdir(exist_ok=True)
    image_list = list(sorted(out_dir.glob('*png'), key=sorter))

    # git_hash = get_git_hash()
    env.filters['basename'] = basename
    template = env.get_template('base.html')

    out_html = out_dir.absolute() / 'eddy_summary.html'
    with open(out_html, 'w') as fh:
        fh.write(template.render(image_list=image_list,
                                 eddyOut=eddyOut,
                                 subject=eddyOut.eddy_prefix,
                                 bwh_fig_loc=bwh_fig_loc,
                                 ))

    replace_image_locations_to_relative_in_html(out_html, out_dir)


def replace_image_locations_to_relative_in_html(
        html_loc:str, image_root: Path) -> None:
    '''Replace image locations to relative location'''
    # Read in the file
    with open(html_loc, 'r') as file :
      filedata = file.readlines()

    # Replace the target string
    new_lines = []
    for line in filedata:
        if 'img src' in line:
            new_line = re.sub(f'{image_root.absolute()}/', '', line)
            new_lines.append(new_line)
        else:
            new_lines.append(line)

    # Write the file out again
    with open(html_loc, 'w') as file:
        for new_line in new_lines:
          file.write(new_line)
    

def get_git_hash() -> str:
    # git version
    command = 'git rev-parse HEAD'
    script_dir = os.path.dirname(os.path.realpath(__file__))
    os.chdir(script_dir)
    git_hash = os.popen(command).read()
    return git_hash
